@@ -40,7 +40,7 @@ struct _XfwWindowWaylandPrivate {
     struct zwlr_foreign_toplevel_handle_v1 *handle;
     gboolean created_emitted;
 
-    guint64 id;
+    const gchar **class_ids;
     gchar *app_id;
     gchar *name;
     XfwWindowState state;
@@ -54,7 +54,8 @@ static void xfw_window_wayland_constructed(GObject *obj);
 static void xfw_window_wayland_set_property(GObject *obj, guint prop_id, const GValue *value, GParamSpec *pspec);
 static void xfw_window_wayland_get_property(GObject *obj, guint prop_id, GValue *value, GParamSpec *pspec);
 static void xfw_window_wayland_finalize(GObject *obj);
-static guint64 xfw_window_wayland_get_id(XfwWindow *window);
+
+static const gchar *const *xfw_window_wayland_get_class_ids(XfwWindow *window);
 static const gchar *xfw_window_wayland_get_name(XfwWindow *window);
 static GIcon *xfw_window_wayland_get_gicon(XfwWindow *window);
 static XfwWindowType xfw_window_wayland_get_window_type(XfwWindow *window);
@@ -115,7 +116,7 @@ xfw_window_wayland_class_init(XfwWindowWaylandClass *klass) {
     gklass->get_property = xfw_window_wayland_get_property;
     gklass->finalize = xfw_window_wayland_finalize;
 
-    window_class->get_id = xfw_window_wayland_get_id;
+    window_class->get_class_ids = xfw_window_wayland_get_class_ids;
     window_class->get_name = xfw_window_wayland_get_name;
     window_class->get_gicon = xfw_window_wayland_get_gicon;
     window_class->get_window_type = xfw_window_wayland_get_window_type;
@@ -155,6 +156,7 @@ xfw_window_wayland_class_init(XfwWindowWaylandClass *klass) {
 static void
 xfw_window_wayland_init(XfwWindowWayland *window) {
     window->priv = xfw_window_wayland_get_instance_private(window);
+    window->priv->class_ids = g_new0(const gchar *, 2);
 }
 
 static void
@@ -198,19 +200,18 @@ xfw_window_wayland_finalize(GObject *obj) {
     XfwWindowWayland *window = XFW_WINDOW_WAYLAND(obj);
 
     zwlr_foreign_toplevel_handle_v1_destroy(window->priv->handle);
+    g_free(window->priv->class_ids);
     g_free(window->priv->app_id);
     g_free(window->priv->name);
     g_list_free(window->priv->monitors);
-    if (window->priv->app) {
-        g_object_unref(window->priv->app);
-    }
+    g_object_unref(window->priv->app);
 
     G_OBJECT_CLASS(xfw_window_wayland_parent_class)->finalize(obj);
 }
 
-static guint64
-xfw_window_wayland_get_id(XfwWindow *window) {
-    return XFW_WINDOW_WAYLAND(window)->priv->id;
+static const gchar *const *
+xfw_window_wayland_get_class_ids(XfwWindow *window) {
+    return XFW_WINDOW_WAYLAND(window)->priv->class_ids;
 }
 
 static const gchar *
@@ -232,7 +233,7 @@ xfw_window_wayland_get_gicon(XfwWindow *window) {
 
 static XfwWindowType
 xfw_window_wayland_get_window_type(XfwWindow *window) {
-    g_message("Window types are not supported on Wayland");
+    _xfw_g_message_once("Window types are not supported on Wayland");
     return XFW_WINDOW_TYPE_NORMAL;
 }
 
@@ -248,7 +249,7 @@ xfw_window_wayland_get_capabilities(XfwWindow *window) {
 
 static GdkRectangle *
 xfw_window_wayland_get_geometry(XfwWindow *window) {
-    g_message("xfw_window_get_geometry() unsupported on Wayland");
+    _xfw_g_message_once("xfw_window_get_geometry() unsupported on Wayland");
     return &XFW_WINDOW_WAYLAND(window)->priv->geometry;
 }
 
@@ -472,10 +473,15 @@ static void
 toplevel_app_id(void *data, struct zwlr_foreign_toplevel_handle_v1 *wl_toplevel, const char *app_id) {
     XfwWindowWayland *window = XFW_WINDOW_WAYLAND(data);
 
+    if (app_id == NULL || *app_id == '\0' || g_strcmp0(app_id, window->priv->app_id) == 0) {
+        return;
+    }
+
     _xfw_window_invalidate_icon(XFW_WINDOW(window));
 
     g_free(window->priv->app_id);
-    window->priv->app_id = (app_id != NULL && *app_id != '\0') ? g_strdup(app_id) : g_strdup("UnknownAppID");
+    window->priv->app_id = g_strdup(app_id);
+    window->priv->class_ids[0] = window->priv->app_id;
 
     if (window->priv->app != NULL) {
         g_object_unref(window->priv->app);
@@ -484,6 +490,8 @@ toplevel_app_id(void *data, struct zwlr_foreign_toplevel_handle_v1 *wl_toplevel,
     g_object_notify(G_OBJECT(window), "application");
 
     g_signal_emit_by_name(window, "icon-changed");
+    g_object_notify(G_OBJECT(window), "class-ids");
+    g_signal_emit_by_name(window, "class-changed");
 }
 
 static void
@@ -527,7 +535,7 @@ toplevel_state(void *data, struct zwlr_foreign_toplevel_handle_v1 *wl_toplevel,
     XfwWindowCapabilities new_capabilities = XFW_WINDOW_CAPABILITIES_NONE;
 
     wl_array_for_each(item, wl_state) {
-        for (size_t i = 0; i < sizeof(state_converters) / sizeof(*state_converters); ++i) {
+        for (size_t i = 0; i < G_N_ELEMENTS(state_converters); ++i) {
             if (state_converters[i].wl_state == *item) {
                 new_state |= state_converters[i].state_bit;
                 break;
@@ -540,7 +548,7 @@ toplevel_state(void *data, struct zwlr_foreign_toplevel_handle_v1 *wl_toplevel,
     g_object_notify(G_OBJECT(window), "state");
     g_signal_emit_by_name(window, "state-changed", changed_mask, new_state);
 
-    for (size_t i = 0; i < sizeof(capabilities_converters) / sizeof(*capabilities_converters); ++i) {
+    for (size_t i = 0; i < G_N_ELEMENTS(capabilities_converters); ++i) {
         if ((new_state & capabilities_converters[i].state_bit) != 0) {
             new_capabilities |= capabilities_converters[i].capabilities_bit_if_present;
         } else {
@@ -619,6 +627,12 @@ toplevel_done(void *data, struct zwlr_foreign_toplevel_handle_v1 *wl_toplevel) {
         XfwScreen *screen = _xfw_window_get_screen(XFW_WINDOW(window));
 
         window->priv->created_emitted = TRUE;
+
+        // nothing in the protocol ensures that the app id is set and we need an app
+        if (window->priv->app == NULL) {
+            toplevel_app_id(window, window->priv->handle, "UnknownAppID");
+        }
+
         g_signal_emit_by_name(screen, "window-opened", window);
         if (window->priv->state & XFW_WINDOW_STATE_ACTIVE) {
             _xfw_screen_wayland_set_active_window(XFW_SCREEN_WAYLAND(screen), XFW_WINDOW(window));
