@@ -30,6 +30,7 @@
 #include "xfw-wnck-icon.h"
 #include "xfw-workspace-x11.h"
 #include "xfw-application-x11.h"
+#include "xfw-x11.h"
 
 enum {
     PROP0,
@@ -39,6 +40,7 @@ enum {
 struct _XfwWindowX11Private {
     WnckWindow *wnck_window;
 
+    const gchar **class_ids;
     XfwWindowType window_type;
     XfwWindowState state;
     XfwWindowCapabilities capabilities;
@@ -53,7 +55,7 @@ static void xfw_window_x11_set_property(GObject *obj, guint prop_id, const GValu
 static void xfw_window_x11_get_property(GObject *obj, guint prop_id, GValue *value, GParamSpec *pspec);
 static void xfw_window_x11_finalize(GObject *obj);
 
-static guint64 xfw_window_x11_get_id(XfwWindow *window);
+static const gchar *const *xfw_window_x11_get_class_ids(XfwWindow *window);
 static const gchar *xfw_window_x11_get_name(XfwWindow *window);
 static GIcon * xfw_window_x11_get_gicon(XfwWindow *window);
 static XfwWindowType xfw_window_x11_get_window_type(XfwWindow *window);
@@ -82,6 +84,7 @@ static gboolean xfw_window_x11_set_below(XfwWindow *window, gboolean is_below, G
 static gboolean xfw_window_x11_is_on_workspace(XfwWindow *window, XfwWorkspace *workspace);
 static gboolean xfw_window_x11_is_in_viewport(XfwWindow *window, XfwWorkspace *workspace);
 
+static void class_changed(WnckWindow *wnck_window, XfwWindowX11 *window);
 static void name_changed(WnckWindow *wnck_window, XfwWindowX11 *window);
 static void icon_changed(WnckWindow *wnck_window, XfwWindowX11 *window);
 static void app_name_changed(XfwApplication *app, GParamSpec *pspec, XfwWindowX11 *window);
@@ -89,6 +92,8 @@ static void type_changed(WnckWindow *wnck_window, XfwWindowX11 *window);
 static void state_changed(WnckWindow *wnck_window, WnckWindowState changed_mask, WnckWindowState new_state, XfwWindowX11 *window);
 static void actions_changed(WnckWindow *wnck_window, WnckWindowActions wnck_changed_mask, WnckWindowActions wnck_new_actions, XfwWindowX11 *window);
 static void geometry_changed(WnckWindow *wnck_window, XfwWindowX11 *window);
+static void monitor_added(GdkDisplay *display, GdkMonitor *monitor, XfwWindowX11 *window);
+static void monitor_removed(GdkDisplay *display, GdkMonitor *monitor, XfwWindowX11 *window);
 static void workspace_changed(WnckWindow *wnck_window, XfwWindowX11 *window);
 
 static XfwWindowType convert_type(WnckWindowType wnck_type);
@@ -109,7 +114,7 @@ xfw_window_x11_class_init(XfwWindowX11Class *klass) {
     gklass->get_property = xfw_window_x11_get_property;
     gklass->finalize = xfw_window_x11_finalize;
 
-    window_class->get_id = xfw_window_x11_get_id;
+    window_class->get_class_ids = xfw_window_x11_get_class_ids;
     window_class->get_name = xfw_window_x11_get_name;
     window_class->get_gicon = xfw_window_x11_get_gicon;
     window_class->get_window_type = xfw_window_x11_get_window_type;
@@ -155,17 +160,37 @@ xfw_window_x11_init(XfwWindowX11 *window) {
 static void xfw_window_x11_constructed(GObject *obj) {
     XfwWindowX11 *window = XFW_WINDOW_X11(obj);
     XfwScreen *screen = _xfw_window_get_screen(XFW_WINDOW(window));
+    GdkDisplay *display = gdk_display_get_default();
+    guint n_monitors = gdk_display_get_n_monitors(display);
+    const gchar *class_name = wnck_window_get_class_group_name(window->priv->wnck_window);
+    const gchar *instance_name = wnck_window_get_class_instance_name(window->priv->wnck_window);
 
+    window->priv->class_ids = g_new0(const gchar *, 3);
+    if (class_name != NULL && *class_name != '\0') {
+        window->priv->class_ids[0] = class_name;
+        window->priv->class_ids[1] = instance_name;
+    } else {
+        window->priv->class_ids[0] = instance_name;
+    }
     window->priv->window_type = convert_type(wnck_window_get_window_type(window->priv->wnck_window));
     window->priv->state = convert_state(window->priv->wnck_window, wnck_window_get_state(window->priv->wnck_window));
     wnck_window_get_geometry(window->priv->wnck_window,
                              &window->priv->geometry.x, &window->priv->geometry.y,
                              &window->priv->geometry.width, &window->priv->geometry.height);
+    for (guint n = 0; n < n_monitors; n++) {
+        GdkMonitor *monitor = gdk_display_get_monitor(display, n);
+        GdkRectangle geom;
+        gdk_monitor_get_geometry(monitor, &geom);
+        if (gdk_rectangle_intersect(&window->priv->geometry, &geom, NULL)) {
+            window->priv->monitors = g_list_prepend(window->priv->monitors, monitor);
+        }
+    }
     window->priv->capabilities = convert_capabilities(window->priv->wnck_window, wnck_window_get_actions(window->priv->wnck_window));
     window->priv->workspace = _xfw_screen_x11_workspace_for_wnck_workspace(XFW_SCREEN_X11(screen),
                                                                            wnck_window_get_workspace(window->priv->wnck_window));
     window->priv->app = XFW_APPLICATION(_xfw_application_x11_get(wnck_window_get_class_group(window->priv->wnck_window), window));
 
+    g_signal_connect(window->priv->wnck_window, "class-changed", G_CALLBACK(class_changed), window);
     g_signal_connect(window->priv->wnck_window, "name-changed", G_CALLBACK(name_changed), window);
     g_signal_connect(window->priv->wnck_window, "icon-changed", G_CALLBACK(icon_changed), window);
     g_signal_connect(window->priv->app, "notify::name", G_CALLBACK(app_name_changed), window);
@@ -173,6 +198,8 @@ static void xfw_window_x11_constructed(GObject *obj) {
     g_signal_connect(window->priv->wnck_window, "state-changed", G_CALLBACK(state_changed), window);
     g_signal_connect(window->priv->wnck_window, "actions-changed", G_CALLBACK(actions_changed), window);
     g_signal_connect(window->priv->wnck_window, "geometry-changed", G_CALLBACK(geometry_changed), window);
+    g_signal_connect(display, "monitor-added", G_CALLBACK(monitor_added), window);
+    g_signal_connect(display, "monitor-removed", G_CALLBACK(monitor_removed), window);
     g_signal_connect(window->priv->wnck_window, "workspace-changed", G_CALLBACK(workspace_changed), window);
 }
 
@@ -209,7 +236,9 @@ xfw_window_x11_get_property(GObject *obj, guint prop_id, GValue *value, GParamSp
 static void
 xfw_window_x11_finalize(GObject *obj) {
     XfwWindowX11 *window = XFW_WINDOW_X11(obj);
+    GdkDisplay *display = gdk_display_get_default();
 
+    g_signal_handlers_disconnect_by_func(window->priv->wnck_window, class_changed, window);
     g_signal_handlers_disconnect_by_func(window->priv->wnck_window, name_changed, window);
     g_signal_handlers_disconnect_by_func(window->priv->wnck_window, icon_changed, window);
     g_signal_handlers_disconnect_by_func(window->priv->app, app_name_changed, window);
@@ -217,8 +246,11 @@ xfw_window_x11_finalize(GObject *obj) {
     g_signal_handlers_disconnect_by_func(window->priv->wnck_window, state_changed, window);
     g_signal_handlers_disconnect_by_func(window->priv->wnck_window, actions_changed, window);
     g_signal_handlers_disconnect_by_func(window->priv->wnck_window, geometry_changed, window);
+    g_signal_handlers_disconnect_by_func(display, monitor_added, window);
+    g_signal_handlers_disconnect_by_func(display, monitor_removed, window);
     g_signal_handlers_disconnect_by_func(window->priv->wnck_window, workspace_changed, window);
 
+    g_free (window->priv->class_ids);
     g_list_free(window->priv->monitors);
     g_object_unref(window->priv->app);
 
@@ -228,9 +260,9 @@ xfw_window_x11_finalize(GObject *obj) {
     G_OBJECT_CLASS(xfw_window_x11_parent_class)->finalize(obj);
 }
 
-static guint64
-xfw_window_x11_get_id(XfwWindow *window) {
-    return wnck_window_get_xid(XFW_WINDOW_X11(window)->priv->wnck_window);
+static const gchar *const *
+xfw_window_x11_get_class_ids(XfwWindow *window) {
+    return XFW_WINDOW_X11(window)->priv->class_ids;
 }
 
 static const gchar *
@@ -275,21 +307,7 @@ xfw_window_x11_get_workspace(XfwWindow *window) {
 
 static GList *
 xfw_window_x11_get_monitors(XfwWindow *window) {
-    XfwWindowX11 *xwindow = XFW_WINDOW_X11(window);
-    GdkMonitor *monitor = NULL;
-    GdkWindow *gwindow = gtk_widget_get_window(GTK_WIDGET(window));
-
-    if (gwindow != NULL) {
-        GdkDisplay *display = gdk_display_get_default();
-
-        monitor = gdk_display_get_monitor_at_window(display, gwindow);
-        if (xwindow->priv->monitors == NULL || monitor != xwindow->priv->monitors->data) {
-            xwindow->priv->monitors = g_list_remove(xwindow->priv->monitors, xwindow->priv->monitors->data);
-            xwindow->priv->monitors = g_list_prepend(xwindow->priv->monitors, monitor);
-        }
-    }
-
-    return xwindow->priv->monitors;
+    return XFW_WINDOW_X11(window)->priv->monitors;
 }
 
 static XfwApplication *
@@ -565,6 +583,21 @@ xfw_window_x11_is_in_viewport(XfwWindow *window, XfwWorkspace *workspace) {
                                       _xfw_workspace_x11_get_wnck_workspace(XFW_WORKSPACE_X11(workspace)));
 }
 
+static void
+class_changed(WnckWindow *wnck_window, XfwWindowX11 *window) {
+    const gchar *class_name = wnck_window_get_class_group_name(wnck_window);
+    const gchar *instance_name = wnck_window_get_class_instance_name(wnck_window);
+    if (class_name != NULL && *class_name != '\0') {
+        window->priv->class_ids[0] = class_name;
+        window->priv->class_ids[1] = instance_name;
+    } else {
+        window->priv->class_ids[0] = instance_name;
+        window->priv->class_ids[1] = NULL;
+    }
+    g_object_notify(G_OBJECT(window), "class-ids");
+    g_signal_emit_by_name(window, "class-changed");
+}
+
 static void
 name_changed(WnckWindow *wnck_window, XfwWindowX11 *window) {
     g_object_notify(G_OBJECT(window), "name");
@@ -622,10 +655,58 @@ actions_changed(WnckWindow *wnck_window, WnckWindowActions wnck_changed_mask, Wn
 
 static void
 geometry_changed(WnckWindow *wnck_window, XfwWindowX11 *window) {
+    GdkDisplay *display = gdk_display_get_default();
+    guint n_monitors = gdk_display_get_n_monitors(display);
+    gboolean notify = FALSE;
+
     wnck_window_get_geometry(wnck_window,
                              &window->priv->geometry.x, &window->priv->geometry.y,
                              &window->priv->geometry.width, &window->priv->geometry.height);
     g_signal_emit_by_name(window, "geometry-changed");
+
+    for (GList *lp = window->priv->monitors; lp != NULL;) {
+        GList *lnext = lp->next;
+        GdkRectangle geom;
+        gdk_monitor_get_geometry(lp->data, &geom);
+        if (!gdk_rectangle_intersect(&window->priv->geometry, &geom, NULL)) {
+            window->priv->monitors = g_list_delete_link(window->priv->monitors, lp);
+            notify = TRUE;
+        }
+        lp = lnext;
+    }
+
+    for (guint n = 0; n < n_monitors; n++) {
+        GdkMonitor *monitor = gdk_display_get_monitor(display, n);
+        GdkRectangle geom;
+        gdk_monitor_get_geometry(monitor, &geom);
+        if (gdk_rectangle_intersect(&window->priv->geometry, &geom, NULL) && !g_list_find(window->priv->monitors, monitor)) {
+            window->priv->monitors = g_list_prepend(window->priv->monitors, monitor);
+            notify = TRUE;
+        }
+    }
+
+    if (notify) {
+        g_object_notify(G_OBJECT(window), "monitors");
+    }
+}
+
+static void
+monitor_added(GdkDisplay *display, GdkMonitor *monitor, XfwWindowX11 *window) {
+    GdkRectangle geom;
+    gdk_monitor_get_geometry(monitor, &geom);
+    if (gdk_rectangle_intersect(&window->priv->geometry, &geom, NULL)) {
+        window->priv->monitors = g_list_prepend(window->priv->monitors, monitor);
+        g_object_notify(G_OBJECT(window), "monitors");
+    }
+}
+
+static void
+monitor_removed(GdkDisplay *display, GdkMonitor *monitor, XfwWindowX11 *window) {
+    GList *lp = g_list_find(window->priv->monitors, monitor);
+    if (lp != NULL) {
+        window->priv->monitors = g_list_delete_link(window->priv->monitors, lp);
+        g_object_notify(G_OBJECT(window), "monitors");
+    }
 }
 
 static void
@@ -688,7 +769,7 @@ static const struct {
 static XfwWindowState
 convert_state(WnckWindow *wnck_window, WnckWindowState wnck_state) {
     XfwWindowState state = XFW_WINDOW_STATE_NONE;
-    for (size_t i = 0; i < sizeof(state_converters) / sizeof(*state_converters); ++i) {
+    for (size_t i = 0; i < G_N_ELEMENTS(state_converters); ++i) {
         if ((wnck_state & state_converters[i].wnck_state_bits) != 0) {
             state |= state_converters[i].state_bit;
         }
@@ -735,7 +816,7 @@ convert_capabilities(WnckWindow *wnck_window, WnckWindowActions wnck_actions)
 {
     WnckWindowState wnck_state = wnck_window_get_state(wnck_window);
     XfwWindowCapabilities capabilities = XFW_WINDOW_CAPABILITIES_NONE;
-    for (size_t i = 0; i < sizeof(capabilities_converters) / sizeof(*capabilities_converters); ++i) {
+    for (size_t i = 0; i < G_N_ELEMENTS(capabilities_converters); ++i) {
         if ((wnck_actions & capabilities_converters[i].wnck_actions_bits) != 0) {
             if ((
                     capabilities_converters[i].need_wnck_state_bits_present
@@ -754,6 +835,26 @@ convert_capabilities(WnckWindow *wnck_window, WnckWindowActions wnck_actions)
     return capabilities;
 }
 
+/**
+ * xfw_window_x11_get_xid:
+ * @window: A #XfwWindow
+ *
+ * On X11, returns the platform-specific #Window handle to the underlying
+ * window.
+ *
+ * It is an error to call this function if the application is not currently
+ * running on X11.
+ *
+ * Return value: An X11 #Window handle.
+ *
+ * Since: 4.19.3
+ **/
+Window
+xfw_window_x11_get_xid(XfwWindow *window) {
+    g_return_val_if_fail(XFW_IS_WINDOW_X11(window), (Window)0);
+    return wnck_window_get_xid(XFW_WINDOW_X11(window)->priv->wnck_window);
+}
+
 WnckWindow *
 _xfw_window_x11_get_wnck_window(XfwWindowX11 *window) {
     return window->priv->wnck_window;
