@@ -26,6 +26,8 @@
 
 #include <glib-object.h>
 
+#include "xfw-workspace-group.h"
+
 G_BEGIN_DECLS
 
 #define XFW_TYPE_WORKSPACE_DUMMY (xfw_workspace_dummy_get_type())
@@ -39,4 +41,8 @@ struct _XfwWorkspaceDummy {
     XfwWorkspaceDummyPrivate *priv;
 };
 
+void _xfw_workspace_dummy_set_workspace_group(XfwWorkspaceDummy *workspace, XfwWorkspaceGroup *group);
+
+G_END_DECLS
+
 #endif  /* __XFW_WORKSPACE_DUMMY_H__ */
